/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_warp.c -- sky and water polygons

#include "quakedef.h"
#include "d3d_quake.h"

extern	model_t	*loadmodel;

int		skytexturenum;

LPDIRECT3DTEXTURE9 solidskytexture = NULL;
LPDIRECT3DTEXTURE9 alphaskytexture = NULL;

msurface_t	*warpface;

extern cvar_t gl_subdivide_size;

void BoundPoly (int numverts, float *verts, vec3_t mins, vec3_t maxs)
{
	int		i, j;
	float	*v;

	mins[0] = mins[1] = mins[2] = 9999;
	maxs[0] = maxs[1] = maxs[2] = -9999;
	v = verts;
	for (i=0 ; i<numverts ; i++)
		for (j=0 ; j<3 ; j++, v++)
		{
			if (*v < mins[j])
				mins[j] = *v;
			if (*v > maxs[j])
				maxs[j] = *v;
		}
}

void SubdividePolygon (int numverts, float *verts)
{
	int		i, j, k;
	vec3_t	mins, maxs;
	float	m;
	float	*v;
	vec3_t	front[64], back[64];
	int		f, b;
	float	dist[64];
	float	frac;
	glpoly_t	*poly;
	float	s, t;

	if (numverts > 60)
		Sys_Error ("numverts = %i", numverts);

	BoundPoly (numverts, verts, mins, maxs);

	for (i=0 ; i<3 ; i++)
	{
		m = (mins[i] + maxs[i]) * 0.5;
		m = gl_subdivide_size.value * floor (m/gl_subdivide_size.value + 0.5);
		if (maxs[i] - m < 8)
			continue;
		if (m - mins[i] < 8)
			continue;

		// cut it
		v = verts + i;
		for (j=0 ; j<numverts ; j++, v+= 3)
			dist[j] = *v - m;

		// wrap cases
		dist[j] = dist[0];
		v-=i;
		VectorCopy (verts, v);

		f = b = 0;
		v = verts;
		for (j=0 ; j<numverts ; j++, v+= 3)
		{
			if (dist[j] >= 0)
			{
				VectorCopy (v, front[f]);
				f++;
			}
			if (dist[j] <= 0)
			{
				VectorCopy (v, back[b]);
				b++;
			}
			if (dist[j] == 0 || dist[j+1] == 0)
				continue;
			if ( (dist[j] > 0) != (dist[j+1] > 0) )
			{
				// clip point
				frac = dist[j] / (dist[j] - dist[j+1]);
				for (k=0 ; k<3 ; k++)
					front[f][k] = back[b][k] = v[k] + frac*(v[3+k] - v[k]);
				f++;
				b++;
			}
		}

		SubdividePolygon (f, front[0]);
		SubdividePolygon (b, back[0]);
		return;
	}

	poly = (glpoly_t *) Hunk_Alloc (sizeof(glpoly_t) + (numverts-4) * VERTEXSIZE*sizeof(float));
	poly->next = warpface->polys;
	warpface->polys = poly;
	poly->numverts = numverts;

	for (i=0 ; i<numverts ; i++, verts+= 3)
	{
		VectorCopy (verts, poly->verts[i]);
		s = DotProduct (verts, warpface->texinfo->vecs[0]);
		t = DotProduct (verts, warpface->texinfo->vecs[1]);

		// store in 5 and 6 so that we can overwrite 3 and 4 at runtime
		poly->verts[i][5] = s;
		poly->verts[i][6] = t;
	}
}

/*
================
GL_SubdivideSurface

Breaks a polygon up along axial 64 unit
boundaries so that turbulent and sky warps
can be done reasonably.
================
*/
void GL_SubdivideSurface (msurface_t *fa)
{
	vec3_t		verts[64];
	int			numverts;
	int			i;
	int			lindex;
	float		*vec;

	warpface = fa;

	//
	// convert edges back to a normal polygon
	//
	numverts = 0;
	for (i=0 ; i<fa->numedges ; i++)
	{
		lindex = loadmodel->surfedges[fa->firstedge + i];

		if (lindex > 0)
			vec = loadmodel->vertexes[loadmodel->edges[lindex].v[0]].position;
		else
			vec = loadmodel->vertexes[loadmodel->edges[-lindex].v[1]].position;
		VectorCopy (vec, verts[numverts]);
		numverts++;
	}

	SubdividePolygon (numverts, verts[0]);
}

//=========================================================



// speed up sin calculations - Ed
float	turbsin[] =
{
	#include "gl_warp_sin.h"
};
#define TURBSCALE (256.0 / (2 * M_PI))

/*
=============
EmitWaterPolys

Does a water warp on the pre-fragmented glpoly_t chain
=============
*/
void EmitWaterPolys (msurface_t *surf)
{
	glpoly_t	*p;
	float		*v;
	int			i;

	for (p = surf->polys; p; p = p->next)
	{
		for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
		{
			// texcoords
			v[3] = (v[5] + turbsin[(int) ((v[6] * 0.125 + realtime) * TURBSCALE) & 255]) * 0.015625f;
			v[4] = (v[6] + turbsin[(int) ((v[5] * 0.125 + realtime) * TURBSCALE) & 255]) * 0.015625f;
		}

		// draw the poly
		d3d_Device->DrawPrimitiveUP (D3DPT_TRIANGLEFAN, p->numverts - 2, p->verts[0], VERTEXSIZE * sizeof (float));
	}
}


/*
=================
R_DrawSkyChain
=================
*/
void R_DrawSkyChain (msurface_t *skychain)
{
	// nosky to draw!
	if (!skychain) return;

	msurface_t *surf;
	float speedscale[2];

	D3D_SetFVFStateManaged (D3DFVF_XYZ | D3DFVF_TEX2);

	D3D_BindTexture (0, solidskytexture);
	D3D_BindTexture (1, alphaskytexture);

	d3d_Device->SetTextureStageState (0, D3DTSS_COLOROP, D3DTOP_MODULATE);
	d3d_Device->SetTextureStageState (0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	d3d_Device->SetTextureStageState (0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
	d3d_Device->SetTextureStageState (0, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

	d3d_Device->SetTextureStageState (1, D3DTSS_COLOROP, D3DTOP_BLENDTEXTUREALPHA);
	d3d_Device->SetTextureStageState (1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	d3d_Device->SetTextureStageState (1, D3DTSS_COLORARG2, D3DTA_CURRENT);

	d3d_Device->SetTextureStageState (2, D3DTSS_COLOROP, D3DTOP_DISABLE);

	speedscale[0] = realtime * 8;
	speedscale[0] -= (int) speedscale[0] & ~127;

	speedscale[1] = realtime * 16;
	speedscale[1] -= (int) speedscale[1] & ~127;

	for (surf = skychain; surf; surf = surf->texturechain)
	{
		glpoly_t	*p;
		float		*v;
		int			i;
		vec3_t	dir;
		float	length;

		for (p = surf->polys; p; p = p->next)
		{
			for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
			{
				VectorSubtract (v, r_origin, dir);
				dir[2] *= 3;	// flatten the sphere

				length = dir[0] * dir[0] + dir[1] * dir[1] + dir[2] * dir[2];
				length = sqrt (length);
				length = 6 * 63 / length;

				dir[0] *= length;
				dir[1] *= length;

				v[3] = (speedscale[0] + dir[0]) * 0.0078125f;
				v[4] = (speedscale[0] + dir[1]) * 0.0078125f;

				v[5] = (speedscale[1] + dir[0]) * 0.0078125f;
				v[6] = (speedscale[1] + dir[1]) * 0.0078125f;
			}

			// draw the poly
			d3d_Device->DrawPrimitiveUP (D3DPT_TRIANGLEFAN, p->numverts - 2, p->verts[0], VERTEXSIZE * sizeof (float));
		}
	}

	d3d_Device->SetTextureStageState (1, D3DTSS_COLOROP, D3DTOP_DISABLE);
}


//===============================================================

/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
==============
*/
void R_InitSky (miptex_t *mt)
{
	int			i, j, p;
	byte		*src;
	unsigned	trans[128 * 128];
	unsigned	transpix;
	int			r, g, b;
	unsigned	*rgba;

	// destroy any current textures we might have
	SAFE_RELEASE (solidskytexture);
	SAFE_RELEASE (alphaskytexture);

	src = (byte *) mt + mt->offsets[0];

	// make an average value for the back to avoid
	// a fringe on the top level
	for (i = 0, r = 0, g = 0, b = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j + 128];

			rgba = &d_8to24table[p];

			trans[(i * 128) + j] = *rgba;

			r += ((byte *) rgba)[0];
			g += ((byte *) rgba)[1];
			b += ((byte *) rgba)[2];
		}
	}

	((byte *) &transpix)[0] = r / (128 * 128);
	((byte *) &transpix)[1] = g / (128 * 128);
	((byte *) &transpix)[2] = b / (128 * 128);
	((byte *) &transpix)[3] = 0;

	// upload it
	D3D_LoadTexture (&solidskytexture, 128, 128, (byte *) trans, NULL, false, false);

	// bottom layer
	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j];

			if (p == 0)
				trans[(i * 128) + j] = transpix;
			else
				trans[(i * 128) + j] = d_8to24table[p];
		}
	}

	// upload it
	D3D_LoadTexture (&alphaskytexture, 128, 128, (byte *) trans, NULL, false, true);
}

